import logging
import os.path
import platform
import re

from edalize.edatool import Edatool

logger = logging.getLogger(__name__)

""" Synopsys (formerly Atrenta) Spyglass Backend

Spyglass performs static source code analysis on HDL code and checks for common
coding errors or coding style violations.

Example snippet of a CAPI2 description file:

spyglass:
  methodology: "GuideWare/latest/block/rtl_handoff"
  goals:
    - lint/lint_rtl
  spyglass_options:
    # prevent error SYNTH_5273 on generic RAM descriptions
    - handlememory yes

"""
class Spyglass(Edatool):

    tool_options = {
        'members' : {
            'methodology' : 'String'
        },
        'lists': {
            'goals': 'String',
            'spyglass_options': 'String',
        },
    }

    argtypes = ['vlogdefine', 'vlogparam']

    tool_options_defaults = {
         'methodology': 'GuideWare/latest/block/rtl_handoff',
         'goals': [ 'lint/lint_rtl' ],
         'spyglass_options': [],
    }

    def _set_tool_options_defaults(self):
        for key, default_value in self.tool_options_defaults.items():
            if not key in self.tool_options:
                logger.info("Set Spyglass tool option %s to default value %s" 
                    % (key, str(default_value)))
                self.tool_options[key] = default_value


    """ Configuration is the first phase of the build

    This writes the project TCL files and Makefile. It first collects all
    sources, IPs and contraints and then writes them to the TCL file along
    with the build steps.
    """
    def configure_main(self):
        self._set_tool_options_defaults()

        (src_files, incdirs) = self._get_fileset_files(force_slash=True)

        self.jinja_env.filters['src_file_filter'] = self.src_file_filter

        has_systemVerilog = False
        for src_file in src_files:
            if src_file.file_type.startswith('systemVerilogSource'):
                has_systemVerilog = True
                break

        template_vars = {
            'name'              : self.name,
            'src_files'         : src_files,
            'incdirs'           : incdirs,
            'tool_options'      : self.tool_options,
            'toplevel'          : self.toplevel,
            'vlogparam'         : self.vlogparam,
            'vlogdefine'        : self.vlogdefine,
            'has_systemVerilog' : has_systemVerilog,
            'sanitized_goals'   : [],
        }

        self.render_template('spyglass-project.prj.j2',
                             self.name + '.prj',
                             template_vars)

        # Create a single TCL file for each goal
        goals = ['Design_Read'] + self.tool_options['goals']

        for goal in goals:
            template_vars['goal'] = goal
            sanitized_goal = re.sub(r"[^a-zA-Z0-9]", '_', goal).lower()
            template_vars['sanitized_goals'].append(sanitized_goal)

            self.render_template('spyglass-run-goal.tcl.j2',
                                 'spyglass-run-%s.tcl' % sanitized_goal,
                                 template_vars)


        self.render_template('Makefile.j2',
                             'Makefile',
                             template_vars)

    def src_file_filter(self, f):
        def _vhdl_source(f):
            s = 'read_file -type vhdl'
            if f.logical_name:
                s += ' -library '+f.logical_name
            return s

        file_types = {
            'verilogSource'       : 'read_file -type verilog',
            'systemVerilogSource' : 'read_file -type verilog',
            'vhdlSource'          : _vhdl_source(f),
            'tclSource'           : 'source',
            'waiver'              : 'read_file -type waiver',
        }
        _file_type = f.file_type.split('-')[0]
        if _file_type in file_types:
            return file_types[_file_type] + ' ' + f.name
        elif _file_type == 'user':
            return ''
        else:
            _s = "{} has unknown file type '{}'"
            logger.warning(_s.format(f.name,
                                     f.file_type))
        return ''



from fusesoc.capi1.core import Core as Capi1Core
from fusesoc.capi2.core import Core as Capi2Core

class Core(object):
    
    def __new__(cls, *args, **kwargs):
        with open(args[0]) as f:
            first_line = f.readline().split()[0]
            if  first_line == 'CAPI=1':
                return Capi1Core(*args, **kwargs)
            elif first_line == 'CAPI=2:':
                return Capi2Core(*args, **kwargs)
            else:
                raise RuntimeError("Unknown file type")



import logging
import os.path
import platform

from edalize.edatool import Edatool

logger = logging.getLogger(__name__)

""" Vivado Backend

The Vivado backend executes Xilinx Vivado to build systems and program the FPGA.

The backend defines the following section:

    [vivado]
    part = <part name> # Format <family><device><package>-<speedgrade>
    hw_device = <device name> # Format <family><device>_0
    top_module = <RTL module name>

A core (usually the system core) can add the following files:

- Standard design sources

- Constraints: Supply xdc files with file_type=xdc

- IP: Supply the IP core xci file with file_type=xci and other files (like .prj)
      as file_type=data
"""
class Vivado(Edatool):

    tool_options = {'members' : {'part' : 'String'}}

    argtypes = ['vlogdefine', 'vlogparam']

    """ Configuration is the first phase of the build

    This writes the project TCL files and Makefile. It first collects all
    sources, IPs and contraints and then writes them to the TCL file along
     with the build steps.
    """
    def configure_main(self):
        (src_files, incdirs) = self._get_fileset_files(force_slash=True)

        self.jinja_env.filters['src_file_filter'] = self.src_file_filter

        has_vhdl2008 = 'vhdlSource-2008' in [x.file_type for x in src_files]
        has_xci      = 'xci'             in [x.file_type for x in src_files]

        template_vars = {
            'name'         : self.name,
            'src_files'    : src_files,
            'incdirs'      : incdirs,
            'tool_options' : self.tool_options,
            'toplevel'     : self.toplevel,
            'vlogparam'    : self.vlogparam,
            'vlogdefine'   : self.vlogdefine,
            'has_vhdl2008' : has_vhdl2008,
            'has_xci'      : has_xci,
        }

        self.render_template('vivado-project.tcl.j2',
                             self.name+'.tcl',
                             template_vars)

        self.render_template('vivado-makefile.j2',
                             'Makefile',
                             {'name' : self.name})

        self.render_template('vivado-run.tcl.j2',
                             self.name+"_run.tcl")

    def src_file_filter(self, f):
        def _vhdl_source(f):
            s = 'read_vhdl'
            if f.file_type == 'vhdlSource-2008':
                s += ' -vhdl2008'
            if f.logical_name:
                s += ' -library '+f.logical_name
            return s

        file_types = {
            'verilogSource'       : 'read_verilog',
            'systemVerilogSource' : 'read_verilog -sv',
            'vhdlSource'          : _vhdl_source(f),
            'xci'                 : 'read_ip',
            'xdc'                 : 'read_xdc',
            'tclSource'           : 'source',
        }
        _file_type = f.file_type.split('-')[0]
        if _file_type in file_types:
            return file_types[_file_type] + ' ' + f.name
        elif _file_type == 'user':
            return ''
        else:
            _s = "{} has unknown file type '{}'"
            logger.warning(_s.format(f.name,
                                     f.file_type))
        return ''

    """ Program the FPGA

    For programming the FPGA a vivado tcl script is written that searches for the
    correct FPGA board and then downloads the bitstream. The tcl script is then
    executed in Vivado's batch mode.
    """
    def run(self, remaining):
        tcl_file_name = self.name+"_pgm.tcl"
        self._write_program_tcl_file(tcl_file_name)
        self._run_tool('vivado', ['-mode', 'batch', '-source', tcl_file_name ])

    """ Write the programming TCL file """
    def _write_program_tcl_file(self, program_tcl_filename):
        template_vars = {}
        template_vars['bitstream_name'] = self.name+'.bit'
        template_vars['hw_device'] = self.tool_options['hw_device']

        template = self.jinja_env.get_template('vivado/vivado-program.tcl.j2')
        tcl_file_path = os.path.join(self.work_root, program_tcl_filename)
        with open(tcl_file_path, 'w') as program_tcl_file:
            program_tcl_file.write(template.render(template_vars))



import os
import pytest
import shutil
import tempfile

from fusesoc.core import Core

from test_common import tests_dir

cores_root = os.path.join(tests_dir, 'cores')

def test_coregen_provider():

    cache_root = tempfile.mkdtemp('coregen_')
    core = Core(os.path.join(cores_root, 'misc', 'coregencore.core'), cache_root)

    os.environ['PATH'] = os.path.join(tests_dir, 'mock_commands')+':'+os.environ['PATH']
    core.setup()

    for f in ['dummy.cgp',
	      'dummy.xco',
	      os.path.join('subdir', 'dummy.extra')]:
        assert(os.path.isfile(os.path.join(core.files_root, f)))

    with open(os.path.join(core.files_root, 'run.cmd')) as f:
        assert(f.read() == '-r -b dummy.xco -p dummy.cgp\n')
    
def test_git_provider():
    cache_root = tempfile.mkdtemp('git_')
    core = Core(os.path.join(cores_root, 'misc', 'gitcore.core'), cache_root)
    
    core.setup()

    for f in ['LICENSE',
	      'README.md',
	      'wb_common.core',
	      'wb_common.v',
	      'wb_common_params.v']:
        assert(os.path.isfile(os.path.join(core.files_root, f)))

def test_github_provider():
    cache_root = tempfile.mkdtemp('github_')
    core = Core(os.path.join(cores_root, 'vlog_tb_utils', 'vlog_tb_utils-1.1.core'), cache_root)
    
    core.setup()
    
    for f in ['LICENSE',
              'vlog_functions.v',
              'vlog_tap_generator.v',
              'vlog_tb_utils.core',
              'vlog_tb_utils.v']:
        assert(os.path.isfile(os.path.join(core.files_root, f)))
        ref_dir   = os.path.join(os.path.dirname(__file__),  __name__)
        f = 'vlog_functions.v'
    with open(os.path.join(ref_dir, f)) as fref, open(os.path.join(core.files_root, f)) as fgen:
            assert fref.read() == fgen.read(), f

def test_logicore_provider():

    cache_root = tempfile.mkdtemp('logicore_')
    core = Core(os.path.join(cores_root, 'misc', 'logicorecore.core'), cache_root)

    os.environ['PATH'] = os.path.join(tests_dir, 'mock_commands')+':'+os.environ['PATH']
    core.setup()

    for f in ['dummy.tcl',
	      'dummy.xci',
	      os.path.join('subdir', 'dummy.extra')]:
        assert(os.path.isfile(os.path.join(core.files_root, f)))

    with open(os.path.join(core.files_root, 'vivado.cmd')) as f:
        assert(f.read() == '-mode batch -source dummy.tcl\n')

def test_opencores_provider():
    cache_root = tempfile.mkdtemp('opencores_')
    core = Core(os.path.join(cores_root, 'misc', 'opencorescore.core'), cache_root)

    core.setup()

    assert(os.path.isfile(os.path.join(core.files_root, 'tap_defines.v')))
    assert(os.path.isfile(os.path.join(core.files_root, 'tap_top.v')))

def test_url_provider():
    cores_root = os.path.join(tests_dir, 'capi2_cores', 'providers')

    for corename in ['url_simple',
                     'url_simple_with_user_agent',
                     'url_tar',
                     'url_zip']:
        cache_root = tempfile.mkdtemp(prefix='url_')
        core = Core(os.path.join(cores_root, corename+'.core'), cache_root)
        core.setup()
        assert(os.path.isfile(os.path.join(core.files_root, 'file.v')))



import pytest

from fusesoc.main import sim
from fusesoc.coremanager import CoreManager
from fusesoc.config import Config

from test_common import cache_root, cores_root, tests_dir

@pytest.mark.xfail
def test_sim(capsys):

    class Args():
        sim = None
        testbench = None
        target = None
        keep = False
        backendargs = None
        setup = True
        build_only = False
        no_export = False
        def __init__(self, system):
            self.system = system

    from fusesoc.config import Config
    from fusesoc.coremanager import CoreManager

    build_root = os.path.join(tests_dir, 'build')
    config = Config()
    config.build_root = build_root
    config.cache_root = cache_root

    common_cm = CoreManager(config)
    common_cm.add_cores_root(cores_root)
    args = Args(system="wb_common")
    with pytest.raises(SystemExit):
        sim(common_cm, args)
    out, err = capsys.readouterr()
    assert out == ""
    #Workaround since this test fails with Travis on Python2.7. No idea why
    import sys
    if sys.version_info[0] > 2:
        assert err == "No tool was supplied on command line or found in 'wb_common' core description\n"



import os
import sys
if sys.version[0] == '2':
    import ConfigParser as configparser
else:
    import configparser

class FusesocConfigParser(configparser.SafeConfigParser):
    def __init__(self, config_file):
        if sys.version[0] == '2':
            configparser.SafeConfigParser.__init__(self)
        else:
            super(FusesocConfigParser, self).__init__()
        if not os.path.exists(config_file):
            raise Exception("Could not find " + config_file)
        f = open(config_file)
        id_string = f.readline().split('=')

        if id_string[0].strip().upper() in ['CAPI', 'SAPI']:
            self.type = id_string[0]
        else:
            raise SyntaxError("Could not find API type in " + config_file)
        try:
            self.version = int(id_string[1].strip())
        except ValueError:
                raise SyntaxError("Unknown version '{}'".format(id_string[1].strip()))

        except IndexError:
            raise SyntaxError("Could not find API version in " + config_file)
        try:
            self.readfp(f)
        except configparser.MissingSectionHeaderError:
            raise SyntaxError("Missing section header")
        except configparser.ParsingError as e:
            raise SyntaxError(e.message)
        except configparser.DuplicateSectionError as e:
            raise SyntaxError(e.message)
        except configparser.DuplicateOptionError as e:
            raise SyntaxError(e.message)

    def get_section(self, section):
        if self.has_section(section):
            return dict(self.items(section))
        else:
            return {}



import argparse
import os

class FileAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, [os.path.abspath(values[0])])

class Plusargs:
    def __init__(self, items):
        self.items = items

    def add_arguments(self, parser):
        for item in self.items:
            tmp = self.items.get(item).split(' ',1)
            _name = '--' + str(item)

            if len(tmp) > 1:
                _help = tmp[1]
            else:
                _help = None
            _type = tmp[0].strip()
            if _type == 'int':
                parser.add_argument(_name, type=int, nargs=1, help=_help)
            elif _type == 'str':
                parser.add_argument(_name, type=str, nargs=1, help=_help)
            elif _type == 'bool':
                parser.add_argument(_name, action='store_true', help=_help)
            elif _type == 'file':
                parser.add_argument(_name, type=str, nargs=1, action=FileAction)
            else:
                raise Exception("Unknown plusargs type '"+_type+"'")



#FIXME: Add IP-XACT support
import importlib
import logging
import os
from pkgutil import walk_packages
from pyparsing import Forward, OneOrMore, Optional, Suppress, Word, alphanums
import shutil
import sys
import yaml

from ipyxact.ipyxact import Component
from fusesoc import utils
from fusesoc.provider import get_provider
from fusesoc.vlnv import Vlnv
from edalize import get_edatools

logger = logging.getLogger(__name__)

class File(object):
    def __init__(self, tree):
        self.copyto = ""
        self.file_type = ''
        self.is_include_file = False
        self.logical_name = ''
        if type(tree) is dict:
            for k, v in tree.items():
                self.name = os.path.expandvars(k)
                self.file_type       = v.get('file_type', '')
                self.is_include_file = v.get('is_include_file', False)
                self.copyto          = v.get('copyto', '')
        else:
            self.name = os.path.expandvars(tree)
            self.is_include_file = False #"FIXME"

class String(str):
    def parse(self, flags):
        _flags = []
        for k,v in flags.items():
            if v == True:
                _flags.append(k)
            elif v in [False, None]:
                pass
            else:
                _flags.append(k+'_'+v)

        def cb_conditional(s,l,t):
            if (t.cond in _flags) != (t.negate == '!'):
                return t.expr
            else:
                return []
        word = Word(alphanums+':<>.[]_-,=~/')
        conditional = Forward()
        conditional << (Optional("!")("negate") + word("cond") + Suppress('?') + Suppress('(') + OneOrMore(conditional ^ word)("expr") + Suppress(')')).setParseAction(cb_conditional)
        #string = (function ^ word)
        string = word
        string_list = OneOrMore(conditional ^ string)
        s = ' '.join(string_list.parseString(self.__str__()))
        logger.debug("Parsing '{}' with flags {} => {}".format(self.__str__(),
                                                               str(_flags),s))
        return s

class StringOrList(object):
    def __new__(cls, *args, **kwargs):
        if type(args[0]) == list:
            return [String(s) for s in args[0]]
        elif type(args[0]) == str:
            return [String(args[0])]

class Section(object):
    members = {}
    lists   = {}
    dicts   = {}
    def __init__(self, tree):
        for k, v in tree.items():
            if not v:
                continue
            if k in self.members:
                setattr(self, k, globals()[self.members[k]](v))
            elif k in self.lists:
                _l = []
                for _item in v:
                    try:
                        _l.append(globals()[self.lists[k]](_item))
                    except TypeError as e:
                        raise SyntaxError("Bad option '{}' in section '{}'".format(_item, k))
                setattr(self, k, _l)
            elif k in self.dicts:
                if not isinstance(v, dict):
                    raise SyntaxError("Object in '{}' section must be a dict".format(k))
                _d = {}
                for _name, _items in v.items():
                    try:
                        _d[_name] = globals()[self.dicts[k]](_items)
                    except AttributeError as e:
                        raise SyntaxError("Bad option '{}' in section '{}'".format(_name, k))
                    setattr(_d[_name], 'name', _name)
                setattr(self, k, _d)
            else:
                raise KeyError(k + " in section " + self.__class__.__name__)

class Provider(object):
    def __new__(cls, *args, **kwargs):
        provider_name = args[0]['name']
        if provider_name is None:
            raise RuntimeError('Missing "name" in section [provider]')
        return get_provider(provider_name)(args[0],
                                           "FIXME: core_root is used by local providers",
                                           "FIXME: cache_root can be set in fetch call")

class Core:
    def __init__(self, core_file, cache_root='', build_root=''):
        basename = os.path.basename(core_file)

        self.core_root = os.path.dirname(core_file)

        try:
            _root = Root(yaml.load(open(core_file)))
        except KeyError as e:
            raise SyntaxError("Unknown item {}".format(e))
        except (yaml.scanner.ScannerError, yaml.constructor.ConstructorError) as e:
            raise SyntaxError(str(e))
        for i in _root.members:
            setattr(self, i, getattr(_root, i))
        for i in _root.lists:
            setattr(self, i, getattr(_root, i))
        for i in _root.dicts:
            setattr(self, i, getattr(_root, i))

        self.export_files = []
        if not self.name:
            raise SyntaxError("Missing 'name' parameter")
        self.sanitized_name = self.name.sanitized_name

        for fs in self.filesets.values():
            if fs.file_type:
                for f in fs.files:
                    if not f.file_type:
                        f.file_type = str(fs.file_type)
            if fs.logical_name:
                for f in fs.files:
                    if not f.logical_name:
                        f.logical_name = str(fs.logical_name)

        if self.provider:
            self.files_root = os.path.join(cache_root,
                                           self.sanitized_name)
            #Ugly hack. Don't like injecting vars
            #How about a setup function or setters?
            self.provider.core_root  = self.core_root
            self.provider.files_root = self.files_root
        else:
            self.files_root = self.core_root

    def cache_status(self):
        if self.provider:
            return self.provider.status()
        else:
            return 'local'

    def export(self, dst_dir, flags={}):
        if os.path.exists(dst_dir):
            shutil.rmtree(dst_dir)

        src_files = [f.name for f in self.get_files(flags)]


        for k, v in self._get_vpi(flags).items():
            src_files += [f.name for f in v['src_files'] + v['inc_files']] #FIXME include files
        self._debug("Exporting {}".format(str(src_files)))

        for scripts in self._get_script_names(flags).values():
            for script in scripts:
                for fs in script.filesets:
                    src_files += [f.name for f in self.filesets[fs].files]

        dirs = list(set(map(os.path.dirname,src_files)))
        for d in dirs:
            if not os.path.exists(os.path.join(dst_dir, d)):
                os.makedirs(os.path.join(dst_dir, d))

        for f in src_files:
            if not os.path.isabs(f):
                if(os.path.exists(os.path.join(self.core_root, f))):
                    shutil.copyfile(os.path.join(self.core_root, f),
                                    os.path.join(dst_dir, f))
                elif (os.path.exists(os.path.join(self.files_root, f))):
                    shutil.copyfile(os.path.join(self.files_root, f),
                                    os.path.join(dst_dir, f))
                else:
                    raise RuntimeError('Cannot find %s in :\n\t%s\n\t%s'
                                  % (f, self.files_root, self.core_root))

    def _get_script_names(self, flags):
        target = self._get_target(flags)
        hooks = {}

        if target and target.hooks:
            for hook in ['pre_build', 'post_build', 'pre_run', 'post_run']:
                scripts = getattr(target.hooks, hook)
                if scripts:
                    hooks[hook] = []
                    for script in self._parse_list(flags, scripts):
                        if not script in self.scripts:
                            raise SyntaxError("Script '{}', requested by target '{}', was not found".format(script, target.name))
                        hooks[hook].append(self.scripts[script])

        return hooks

    def get_scripts(self, files_root, flags):
        self._debug("Getting hooks for flags '{}'".format(str(flags)))
        hooks = {}

        for hook, scripts in self._get_script_names(flags).items():
            hooks[hook] = []
            for script in scripts:
                env = script.env
                env['FILES_ROOT'] = files_root
                _script = {'name' : script.name,
                           'cmd'  : [str(x) for x in script.cmd],
                           'env'  : env}
                hooks[hook].append(_script)
                _s = " Matched {} hook {}"
                self._debug(_s.format(hook, str(_script)))
        return hooks

    def get_tool(self, flags):
        self._debug("Getting tool for flags {}".format(str(flags)))
        tool = None
        if flags['tool']:
            tool = flags['tool']
        else:
            _flags = flags.copy()
            _flags['is_toplevel'] = True
            target = self._get_target(_flags)
            if target and target.default_tool:
                tool = target.default_tool

        if tool:
            self._debug(" Matched tool {}".format(tool))
        else:
            self._debug(" Matched no tool")
        return str(tool)

    def get_tool_options(self, flags):
        _flags = flags.copy()

        self._debug("Getting tool options for flags {}".format(str(_flags)))
        target = self._get_target(_flags)
        section = None
        try:
            section = getattr(target.tools, flags['tool'])
        except AttributeError:
            pass

        options = {}
        if section:
            for member in section.members:
                if hasattr(section, member):
                    _member = getattr(section, member)
                    if _member:
                        options[member] = str(_member)
            for member in section.lists:
                if hasattr(section, member):
                    _member = getattr(section, member)
                    if _member:
                        options[member] = [str(x) for x in _member]
        self._debug("Found tool options {}".format(str(options)))
        return options

    def get_depends(self, flags): #Add use flags?
        depends = []
        self._debug("Getting dependencies for flags {}".format(str(flags)))
        for fs in self._get_filesets(flags):
            depends += [Vlnv(d) for d in self._parse_list(flags, fs.depend)]
        return depends

    def get_files(self, flags):
        src_files = []
        for fs in self._get_filesets(flags):
            src_files += fs.files
        return src_files

    def get_parameters(self, flags={}):
        self._debug("Getting parameters for flags '{}'".format(str(flags)))
        target = self._get_target(flags)
        parameters = {}

        if target:
            for _param in target.parameters:
                plist = _param.parse(flags).split('=', 1)

                p = plist[0]

                if not p:
                    continue

                if not p in self.parameters:
                    raise SyntaxError("Parameter '{}', requested by target '{}', was not found".format(p, target.name))

                datatype    = self.parameters[p].datatype
                if len(plist) > 1:
                    default = plist[1]
                else:
                    default     = self.parameters[p].default
                description = self.parameters[p].description
                paramtype   = self.parameters[p].paramtype

                if not datatype in ['bool', 'file', 'int', 'str']:
                    _s = "{} : Invalid datatype '{}' for parameter {}"
                    raise SyntaxError(_s.format(self.name, datatype, p))

                if not paramtype in ['cmdlinearg', 'generic', 'plusarg',
                                     'vlogdefine', 'vlogparam']:
                    _s = "{} : Invalid paramtype '{}' for parameter {}"
                    raise SyntaxError(_s.format(self.name, paramtype, p))
                parameters[p] = {
                    'datatype'  : self.parameters[p].datatype,
                    'paramtype' : self.parameters[p].paramtype,
                }

                if description:
                    parameters[p]['description'] = str(description)

                if default:
                    if datatype == 'bool':
                        if default.lower() ==  'true':
                            parameters[p]['default'] = True
                        elif default.lower() == 'false':
                            parameters[p]['default'] = False
                        else:
                            _s = "{}: Invalid default value '{}' for bool parameter {}"
                            raise SyntaxError(_s.format(self.name, default, p))
                    elif datatype == 'int':
                        if type(default) == int:
                            parameters[p]['default'] = default
                        else:
                            parameters[p]['default'] = int(default,0)
                    else:
                        parameters[p]['default'] = str(default)
        self._debug("Found parameters {}".format(parameters))
        return parameters

    def get_toplevel(self, flags):
        _flags = flags.copy()
        _flags['is_toplevel'] = True #FIXME: Is this correct?
        self._debug("Getting toplevel for flags {}".format(str(_flags)))
        target = self._get_target(_flags)
        if target.toplevel:
            toplevel = self._parse_list(_flags, target.toplevel)
            self._debug("Matched toplevel {}".format(toplevel))
            return ' '.join(toplevel)
        else:
            s = "{} : Target '{}' has no toplevel"
            raise SyntaxError(s.format(self.name, target.name))

    def get_work_root(self, flags):
        _flags = flags.copy()
        _flags['is_toplevel'] = True
        target = self._get_target(_flags)
        if target:
            _flags['target'] = target.name
            tool = self.get_tool(_flags)
            if tool:
                return target.name + '-' + tool
            else:
                raise SyntaxError("Failed to determine work root. Could not resolve tool for target " + target.name)
        else:
            raise SyntaxError("Failed to determine work root. Could not resolve target")

    def _get_vpi(self, flags):
        vpi = {}
        target = self._get_target(flags)
        if not target:
            return vpi
        for vpi_name in self._parse_list(flags, target.vpi):
            vpi_lib = self.vpi[vpi_name]
            files = []
            incfiles = [] #Really do this automatically?
            for fs in vpi_lib.filesets:
                for f in self.filesets[fs].files:
                    if f.is_include_file:
                        incfiles.append(f)
                    else:
                        files.append(f)
            vpi[vpi_name] = {'src_files' : files,
                             'inc_files' : incfiles,
                             'libs'      : [str(l) for l in vpi_lib.libs]}
        return vpi

    def get_vpi(self, flags):
        self._debug("Getting VPI libraries for flags {}".format(flags))
        target = self._get_target(flags)
        vpi = []
        _vpi = self._get_vpi(flags)
        self._debug(" Matched VPI libraries {}".format([v for v in _vpi]))
        for k, v in sorted(_vpi.items()):
            vpi.append({'name'         : k,
                        'src_files'    : [f.name for f in v['src_files']],
                        'include_dirs' : utils.unique_dirs(v['inc_files']),
                        'libs'         : v['libs'],
            })
        return vpi

    def info(self):
        s = """CORE INFO
Name:      {}
Core root: {}

Targets:
{}
"""
        targets = '\n'.join(sorted(self.targets))
        return s.format(str(self.name),
                        str(self.core_root),
                        targets)

    def patch(self, dst_dir):
        #FIXME: Use native python patch instead
        patches = self.provider.patches
        for f in patches:
            patch_file = os.path.abspath(os.path.join(self.core_root, f))
            if os.path.isfile(patch_file):
                self._debug("  applying patch file: " + patch_file + "\n" +
                             "                   to: " + os.path.join(dst_dir))
                try:
                    utils.Launcher('git', ['apply', '--unsafe-paths',
                                     '--directory', os.path.join(dst_dir),
                                     patch_file]).run()
                except OSError:
                    print("Error: Failed to call external command 'patch'")
                    return False
        return True

    def setup(self):
        if self.provider:
            if self.provider.fetch():
                self.patch(self.files_root)

    def _debug(self, msg):
        logger.debug("{} : {}".format(str(self.name), msg))

    def _get_target(self, flags):
        self._debug(" Resolving target for flags '{}'".format(str(flags)))

        target_name = None
        if flags.get('is_toplevel') and flags.get('target'):
            target_name = flags.get('target')
        else:
            target_name = "default"

        if target_name in self.targets:
            self._debug(" Matched target {}".format(target_name))
            return self.targets[target_name]
        else:
            self._debug("Matched no target")

    def _get_filesets(self, flags):
        self._debug("Getting filesets for flags '{}'".format(str(flags)))
        target = self._get_target(flags)
        if not target:
            return []
        filesets = []

        for fs in self._parse_list(flags, target.filesets):
            if not fs in self.filesets:
                raise SyntaxError("Fileset '{}', requested by fileset '{}', was not found".format(fs, target.name))
            filesets.append(self.filesets[fs])

        self._debug(" Matched filesets {}".format(target.filesets))
        return filesets

    def _parse_list(self, flags, l):
        r = []
        for x in l:
            _x = x.parse(flags)
            if _x:
                r.append(_x)
        return r
    #return [x.parse(flags) for x in l if x.parse(flags)]
description = """
---
Root:
  members:
    name        : Vlnv
    description : String
    provider    : Provider
    CAPI=2      : String
  dicts:
    filesets   : Fileset
    scripts    : Script
    targets    : Target
    parameters : Parameter
    vpi        : Vpi

Fileset:
  members:
    file_type    : String
    logical_name : String
  lists:
    files      : File
    depend     : String

Target:
  members:
    default_tool : String
    hooks      : Hooks
    tools    : Tools
    toplevel   : StringOrList
  lists:
    filesets   : String
    flags      : String #FIXME
    parameters : String
    vpi        : String

Tools:
  members : {}
Hooks:
  lists:
    pre_build  : String
    post_build : String
    pre_run    : String
    post_run   : String

Parameter:
  members:
    datatype : String
    default  : String
    description : String
    paramtype   : String
    scope       : String

Script:
  lists:
    cmd      : String
    filesets : String
  dicts:
    env : String

Vpi:
  lists:
    libs         : String
    filesets : String

"""

def _generate_classes(j, base_class):
    for cls, _items in j.items():
        class_members = {}
        if 'members' in _items:
            for key in _items['members']:
                class_members[key] = None
            class_members['members'] = _items['members']
        if 'lists' in _items:
            for key in _items['lists']:
                class_members[key] = []
            class_members['lists'] = _items['lists']
        if 'dicts' in _items:
            for key in _items['dicts']:
                class_members[key] = {}
            class_members['dicts'] = _items['dicts']

        generatedClass = type(cls, (base_class,), class_members)
        globals()[generatedClass.__name__] = generatedClass

capi2_data = yaml.load(description)

for backend in get_edatools():
    if hasattr(backend, 'tool_options'):
        backend_name = backend.__name__
        tool_options = getattr(backend, 'tool_options')
        capi2_data['Tools']['members'][backend_name.lower()] = backend_name
        capi2_data[backend_name] = tool_options

_generate_classes(capi2_data, Section)



import logging
import shutil
import os.path
import subprocess
from fusesoc.provider.provider import Provider
from fusesoc.utils import Launcher

logger = logging.getLogger(__name__)

class Git(Provider):
    @staticmethod
    def init_library(library):
        logger.info("Cloning library into {}".format(library['location']))
        git_args = ['clone', library['sync-uri'], library['location']]
        try:
            Launcher('git', git_args).run()
        except subprocess.CalledProcessError as e:
            raise RuntimeError(str(e))

    @staticmethod
    def update_library(library):
        logger.info("Updating library {}".format(library['location']))
        git_args = ['-C', library['location'], 'pull']
        try:
            Launcher('git', git_args).run()
        except subprocess.CalledProcessError as e:
            raise RuntimeError(str(e))

    def _checkout(self, local_dir):
        version = self.config.get('version', None)

        #TODO : Sanitize URL
        repo   = self.config.get('repo')
        logger.info("Checking out " + repo + " to " + local_dir)
        args = ['clone', '-q', repo, local_dir]
        Launcher('git', args).run()
        if version:
            args = ['-C', local_dir, 'checkout', '-q', version]
            Launcher('git', args).run()



#!/usr/bin/env python
import argparse
import os
import subprocess
import sys
import signal

from fusesoc import __version__

#Check if this is run from a local installation
fusesocdir = os.path.abspath(os.path.join(os.path.dirname(os.path.realpath(__file__)), ".."))
if os.path.exists(os.path.join(fusesocdir, "fusesoc")):
    sys.path[0:0] = [fusesocdir]

from fusesoc.config import Config
from fusesoc.coremanager import CoreManager, DependencyError
from fusesoc.edalizer import Edalizer
from edalize import get_edatool
from fusesoc.provider import get_provider
from fusesoc.vlnv import Vlnv
from fusesoc.utils import Launcher, setup_logging

import logging

logger = logging.getLogger(__name__)

REPOS = [('orpsoc-cores',
          'https://github.com/openrisc/orpsoc-cores',
          "old base library"),
         ('fusesoc-cores',
          'https://github.com/fusesoc/fusesoc-cores',
          "new base library")]

def _get_core(cm, name):
    core = None
    try:
        core = cm.get_core(Vlnv(name))
    except RuntimeError as e:
        logger.error(str(e))
        exit(1)
    except DependencyError as e:
        logger.error("'" + name + "' or any of its dependencies requires '" + e.value + "', but this core was not found")
        exit(1)
    return core

def abort_handler(signal, frame):
        print('');
        logger.info('****************************')
        logger.info('****   FuseSoC aborted  ****')
        logger.info('****************************')
        print('');
        sys.exit(0)

signal.signal(signal.SIGINT, abort_handler)

def build(cm, args):
    do_configure = True
    do_build = not args.setup
    do_run = False
    if not args.target:
        target = 'synth'
    else:
        target = args.target
    flags = {'target' : target,
             'tool' : args.tool}
    run_backend(cm, not args.no_export,
                do_configure, do_build, do_run,
                flags, args.system, args.backendargs)

def pgm(cm, args):
    do_configure = False
    do_build = False
    do_run = True
    flags = {'target' : 'synth',
             'tool' : None}
    run_backend(cm, 'build',
                do_configure, do_build, do_run,
                flags, args.system, args.backendargs)

def fetch(cm, args):
    core = _get_core(cm, args.core)

    try:
        core.setup()
    except RuntimeError as e:
        logger.error("Failed to fetch '{}': {}".format(core.name, str(e)))
        exit(1)

def init(cm, args):
    # Fix Python 2.x.
    global input
    try:
        input = raw_input
    except NameError:
        pass

    xdg_config_home = os.environ.get('XDG_CONFIG_HOME') or \
                      os.path.join(os.path.expanduser('~'), '.config')
    config_file = os.path.join(xdg_config_home, 'fusesoc', 'fusesoc.conf')


    if os.path.exists(config_file):
        logger.warning("'{}' already exists".format(config_file))
        #TODO. Prepend cores_root to file if it doesn't exist
        f = open(config_file, 'w+')
    else:
        logger.info("Writing configuration file to '{}'".format(config_file))
        if not os.path.exists(os.path.dirname(config_file)):
            os.makedirs(os.path.dirname(config_file))
        f = open(config_file,'w+')

    config = Config(file=f)

    xdg_data_home = os.environ.get('XDG_DATA_HOME') or \
             os.path.join(os.path.expanduser('~'), '.local/share')
    _repo_paths = []
    for repo in REPOS:
        name = repo[0]
        library = {
                'sync-uri': repo[1],
                'sync-type': 'git'
                }

        default_dir = os.path.join(xdg_data_home, name)
        prompt = 'Directory to use for {} ({}) [{}] : '
        if args.y:
            location = None
        else:
            location = input(prompt.format(repo[0], repo[2], default_dir))
        if location:
            library['location'] = location
        else:
            location = default_dir
        if os.path.exists(location):
            logger.warning("'{}' already exists".format(location))
            #TODO: Prompt for overwrite
        else:
            logger.info("Initializing {}".format(name))
            try:
                config.add_library(name, library)
            except RuntimeError as e:
                logger.error("Init failed: " + str(e))
                exit(1)
    logger.info("FuseSoC is ready to use!")

def list_paths(cm, args):
    cores_root = cm.get_cores_root()
    print("\n".join(cores_root))

def add_library(cm, args):
    library = {}
    name = args.name
    sync_uri = vars(args)['sync-uri']
    if 'sync-type' in vars(args):
        provider = vars(args)['sync-type']
        library['sync-type'] = provider
    elif os.path.isdir(sync_uri):
        provider = 'local'
        library['sync-type'] = provider
    else:
        provider = 'git'

    if provider == 'local' and not args.location:
        logger.info("Interpreting sync-uri '{}' as location for local provider.".format(sync_uri))
        library['location'] = os.path.abspath(sync_uri)
        library['sync-type'] = 'local'
    elif provider == 'local' and os.path.abspath(args.location) != os.path.abspath(sync_uri):
        logger.error("Location '{}' does not match sync-uri '{}' for local provider. Consider specifying only sync-uri.".format(args.location, sync_uri))
        exit(1)
    else:
        library['sync-uri'] = sync_uri
        if args.location:
            library['location'] = os.path.abspath(args.location)
    if args.no_auto_sync:
        library['auto-sync'] = False

    if args.config:
        config = Config(file=args.config)
    elif vars(args)['global']:
        xdg_config_home = os.environ.get('XDG_CONFIG_HOME') or \
                          os.path.join(os.path.expanduser('~'), '.config')
        config_file = os.path.join(xdg_config_home, 'fusesoc', 'fusesoc.conf')
        config = Config(path=config_file)
    else:
        config = Config(path="fusesoc.conf")

    try:
        config.add_library(name, library)
    except RuntimeError as e:
        logger.error("`add library` failed: " + str(e))
        exit(1)


def list_cores(cm, args):
    cores = cm.get_cores()
    print("\nAvailable cores:\n")
    if not cores:
        cores_root = cm.get_cores_root()
        if cores_root:
            logger.error("No cores found in "+':'.join(cores_root))
        else:
            logger.error("cores_root is not defined")
        exit(1)
    maxlen = max(map(len,cores.keys()))
    print('Core'.ljust(maxlen) + '   Cache status')
    print("="*80)
    for name in sorted(cores.keys()):
        core = cores[name]
        print(name.ljust(maxlen) + ' : ' + core.cache_status())

def core_info(cm, args):
    core = _get_core(cm, args.core)
    print(core.info())

def run(cm, args):
    stages = (args.setup, args.build, args.run)
    #Run all stages by default if no stage flags are set
    if  stages == (False, False, False):
        do_configure = True
        do_build     = True
        do_run       = True
    elif stages == (True, False, True):
        logger.error("Configure and run without build is invalid")
        exit(1)
    else:
        do_configure = args.setup
        do_build     = args.build
        do_run       = args.run

    flags = {'tool'   : args.tool,
             'target' : args.target}
    run_backend(cm,
                not args.no_export,
                do_configure, do_build, do_run,
                flags, args.system, args.backendargs)


def run_backend(cm, export, do_configure, do_build, do_run, flags, system, backendargs):
    tool_error = "No tool was supplied on command line or found in '{}' core description"
    core = _get_core(cm, system)
    try:
        tool = core.get_tool(flags)
    except SyntaxError as e:
        logger.error(str(e))
        exit(1)
    if not tool:
        logger.error(tool_error.format(system))
        exit(1)
    flags['tool'] = tool
    if export:
        export_root = os.path.join(cm.config.build_root, core.name.sanitized_name, 'src')
    else:
        export_root = None
    try:
        work_root   = os.path.join(cm.config.build_root,
                                   core.name.sanitized_name,
                                   core.get_work_root(flags))
    except SyntaxError as e:
        logger.error(e.msg)
        exit(1)
    eda_api_file = os.path.join(work_root,
                                core.name.sanitized_name+'.eda.yml')
    if do_configure:
        try:
            cores = cm.get_depends(core.name, flags)
        except DependencyError as e:
            logger.error(e.msg + "\nFailed to resolve dependencies for {}".format(system))
            exit(1)
        try:
            edalizer = Edalizer(core.name,
                                flags,
                                cores,
                                work_root=work_root,
                                export_root=export_root)
        except SyntaxError as e:
            logger.error(e.msg)
            exit(1)
        except RuntimeError as e:
            logger.error("Setup failed : {}".format(str(e)))
            exit(1)
        edalizer.to_yaml(eda_api_file)

    #Frontend/backend separation

    try:
        backend = get_edatool(tool)(eda_api_file=eda_api_file)
    except ImportError:
        logger.error('Backend "{}" not found'.format(tool))
        exit(1)
    except RuntimeError as e:
        logger.error(str(e))
        exit(1)
    except FileNotFoundError as e:
        logger.error('Could not find EDA API file "{}"'.format(e.filename))
        exit(1)

    if do_configure:
        try:
            backend.configure(backendargs)
            print('')
        except RuntimeError as e:
            logger.error("Failed to configure the system")
            logger.error(str(e))
            exit(1)

    if do_build:
        try:
            backend.build()
        except RuntimeError as e:
            logger.error("Failed to build {} : {}".format(str(core.name),
                                                          str(e)))
            exit(1)

    if do_run:
        try:
            backend.run(backendargs)
        except RuntimeError as e:
            logger.error("Failed to run {} : {}".format(str(core.name),
                                                        str(e)))
            exit(1)

def sim(cm, args):
    do_configure = not args.keep
    do_build = not (args.setup or args.keep)
    do_run   = not (args.build_only or args.setup)

    flags = {'flow' : 'sim',
             'tool' : args.sim,
             'target' : 'sim',
             'testbench' : args.testbench
    }
    run_backend(cm, not args.no_export,
                do_configure, do_build, do_run,
                flags, args.system, args.backendargs)

def update(cm, args):
    libraries = args.libraries
    for root in cm.get_cores_root():
        if not root in cm.config.cores_root:
            # This is a library - handled differently
            continue
        if os.path.exists(root) and (not libraries or root in libraries):
            args = ['-C', root,
                    'config', '--get', 'remote.origin.url']
            repo_root = ""
            try:
                repo_root = subprocess.check_output(['git'] + args).decode("utf-8")
                if repo_root.strip() in [repo[1] for repo in REPOS]:
                    logger.info("Updating '{}'".format(root))
                    args = ['-C', root, 'pull']
                    Launcher('git', args).run()
            except subprocess.CalledProcessError:
                pass

    for (name, library) in cm.config.libraries.items():
        # TODO I don't like this huge "if", get rid of it
        if os.path.exists(library['location']) and \
                not library['sync-uri'] is None and \
                (name in libraries or \
                library['location'] in libraries or \
                library['auto-sync']):
            logger.info("Updating '{}'".format(name))
            try:
                provider = get_provider(library['sync-type'])
            except ImportError as e:
                logger.error("Invalid sync-type '{}' for library '{}' - skipping".format(library['sync-type'], name))
                continue
            try:
                provider.update_library(library)
            except RuntimeError as e:
                logger.error("Failed to update library: " + str(e) + ". Continuing...")

def init_logging(verbose, monochrome, log_file=None):
    level = logging.DEBUG if verbose else logging.INFO

    setup_logging(level, monochrome, log_file)

    if verbose:
        logger.debug("Verbose output")
    else:
        logger.debug("Concise output")

    if monochrome:
        logger.debug("Monochrome output")
    else:
        logger.debug("Colorful output")

def init_coremanager(config, args_cores_root):
    logger.debug("Command line arguments: " + str(sys.argv))

    if os.getenv("FUSESOC_CORES"):
        logger.debug("FUSESOC_CORES: " + str(os.getenv("FUSESOC_CORES").split(':')))
    cm = CoreManager(config)

    # Get the environment variable for further cores
    env_cores_root = []
    if os.getenv("FUSESOC_CORES"):
        env_cores_root = os.getenv("FUSESOC_CORES").split(":")
    env_cores_root.reverse()

    core_libraries = [l['location'] for l in config.libraries.values()]
    for cores_root in config.cores_root + \
                       [config.systems_root] + \
                       env_cores_root + \
                       core_libraries + \
                       args_cores_root:
        try:
            cm.add_cores_root(cores_root)
        except (RuntimeError, IOError) as e:
            logger.warning("Failed to register cores root '{}'".format(str(e)))

    return cm

def parse_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    # Global actions
    parser.add_argument('--version', help='Display the FuseSoC version', action='version', version=__version__)

    # Global options
    parser.add_argument('--cores-root', help='Add additional directories containing cores', default=[], action='append')
    parser.add_argument('--config', help='Specify the config file to use', type=argparse.FileType('r'))
    parser.add_argument('--monochrome', help='Don\'t use color for messages', action='store_true')
    parser.add_argument('--verbose', help='More info messages', action='store_true')
    parser.add_argument('--log-file', help='Write log messages to file')

    # build subparser
    parser_build = subparsers.add_parser('build', help='Build an FPGA load module')
    parser_build.add_argument('--no-export', action='store_true', help='Reference source files from their current location instead of exporting to a build tree')
    parser_build.add_argument('--setup', action='store_true', help='Only create the project files without running the EDA tool')
    parser_build.add_argument('--target', help='Override default target')
    parser_build.add_argument('--tool', help='Override default tool for target')
    parser_build.add_argument('system')
    parser_build.add_argument('backendargs', nargs=argparse.REMAINDER)
    parser_build.set_defaults(func=build)

    # init subparser
    parser_init = subparsers.add_parser('init', help='Initialize the FuseSoC core libraries')
    parser_init.add_argument('-y', action='store_true', help='Skip user input and use default settings')
    parser_init.set_defaults(func=init)

    # pgm subparser
    parser_pgm = subparsers.add_parser('pgm', help='Program an FPGA with a system configuration')
    parser_pgm.add_argument('system')
    parser_pgm.add_argument('backendargs', nargs=argparse.REMAINDER)
    parser_pgm.set_defaults(func=pgm)

    # fetch subparser
    parser_fetch = subparsers.add_parser('fetch', help='Fetch a remote core and its dependencies to local cache')
    parser_fetch.add_argument('core')
    parser_fetch.set_defaults(func=fetch)

    # list-cores subparser
    parser_list_cores = subparsers.add_parser('list-cores', help='List available cores')
    parser_list_cores.set_defaults(func=list_cores)

    # core-info subparser
    parser_core_info = subparsers.add_parser('core-info', help='Display details about a core')
    parser_core_info.add_argument('core')
    parser_core_info.set_defaults(func=core_info)

    # list-paths subparser
    parser_list_paths = subparsers.add_parser('list-paths', help='Display the search order for core root paths')
    parser_list_paths.set_defaults(func=list_paths)

    # library subparser
    parser_library = subparsers.add_parser('library', help='Subcommands for dealing with library management')
    library_subparsers = parser_library.add_subparsers()

    # library add subparser
    parser_library_add = library_subparsers.add_parser('add', help='Add new library to fusesoc.conf')
    parser_library_add.add_argument('name', help='A friendly name  for the library')
    parser_library_add.add_argument('sync-uri', help='The URI source for the library (can be a file system path)')
    parser_library_add.add_argument('--location', help='The location to store the library into (defaults to $XDG_DATA_HOME/[name])')
    parser_library_add.add_argument('--sync-type', help="The provider type for the library. Defaults to 'git'.", choices=['git', 'local'])
    parser_library_add.add_argument('--no-auto-sync', action='store_true', help='Disable automatic updates of the library')
    parser_library_add.add_argument('--global', action='store_true', help='Use the global FuseSoc config file in $XDG_CONFIG_HOME/fusesoc/fusesoc.conf')
    parser_library_add.set_defaults(func=add_library)

    # run subparser
    parser_run = subparsers.add_parser('run', help="Start a tool flow")
    parser_run.add_argument('--no-export', action='store_true', help='Reference source files from their current location instead of exporting to a build tree')
    parser_run.add_argument('--setup',  action='store_true', help="Execute setup stage")
    parser_run.add_argument('--build',  action='store_true', help="Execute build stage")
    parser_run.add_argument('--run'  ,  action='store_true', help="Execute run stage")
    parser_run.add_argument('--target', help='Override default target')
    parser_run.add_argument('--tool', help="Override default tool for target")
    parser_run.add_argument('system', help='Select a system to operate on')
    parser_run.add_argument('backendargs', nargs=argparse.REMAINDER, help="arguments to be sent to backend")
    parser_run.set_defaults(func=run)

    # sim subparser
    parser_sim = subparsers.add_parser('sim', help='Setup and run a simulation')
    parser_sim.add_argument('--no-export', action='store_true', help='Reference source files from their current location instead of exporting to a build tree')
    parser_sim.add_argument('--sim', help='Override the simulator settings from the system file')
    parser_sim.add_argument('--setup', action='store_true', help='Only create the project files without running the EDA tool')
    parser_sim.add_argument('--build-only', action='store_true', help='Build the simulation binary without running the simulator')
    parser_sim.add_argument('--force', action='store_true', help='Force rebuilding simulation model when directory exists')
    parser_sim.add_argument('--keep', action='store_true', help='Prevent rebuilding simulation model if it exists')
    parser_sim.add_argument('--target', help='Override default target')
    parser_sim.add_argument('--testbench', help='Override default testbench')
    parser_sim.add_argument('system', help='Select a system to simulate')
    parser_sim.add_argument('backendargs', nargs=argparse.REMAINDER)
    parser_sim.set_defaults(func=sim)

    # update subparser
    parser_update = subparsers.add_parser('update', help='Update the FuseSoC core libraries')
    parser_update.add_argument('libraries', nargs='*', help='The libraries (or core roots) to update (defaults to all)')
    parser_update.set_defaults(func=update)

    args = parser.parse_args()

    if hasattr(args, 'func'):
        return args
    else:
        parser.print_help()
        return None

def main():

    args = parse_args()
    if not args:
        exit(0)

    init_logging(args.verbose, args.monochrome, args.log_file)
    config = Config(file=args.config)
    cm = init_coremanager(config, args.cores_root)

    # Run the function
    args.func(cm, args)

if __name__ == "__main__":
    main()



import os
import logging

from edalize.edatool import Edatool

logger = logging.getLogger(__name__)

class Icarus(Edatool):

    tool_options = {
        'members' : {'timescale' : 'String'},
        'lists' : {'iverilog_options' : 'String'}
    }

    argtypes = ['plusarg', 'vlogdefine', 'vlogparam']

    MAKEFILE_TEMPLATE = """
all: $(VPI_MODULES) $(TARGET)

$(TARGET):
	iverilog -s$(TOPLEVEL) -c $(TARGET).scr -o $@ $(IVERILOG_OPTIONS)

run: $(VPI_MODULES) $(TARGET)
	vvp -n -M. -l icarus.log -lxt2 $(patsubst %.vpi,-m%,$(VPI_MODULES)) $(TARGET) $(EXTRA_OPTIONS)

clean:
	$(RM) $(VPI_MODULES) $(TARGET)
"""

    VPI_MAKE_SECTION = """
{name}_LIBS := {libs}
{name}_INCS := {incs}
{name}_SRCS := {srcs}

{name}.vpi: $({name}_SRCS)
	iverilog-vpi --name={name} $({name}_LIBS) $({name}_INCS) $?

clean_{name}:
	$(RM) {name}.vpi
"""

    def configure_main(self):
        f = open(os.path.join(self.work_root, self.name+'.scr'),'w')

        (src_files, incdirs) = self._get_fileset_files()
        for key, value in self.vlogdefine.items():
            f.write('+define+{}={}\n'.format(key, self._param_value_str(value, '')))

        for key, value in self.vlogparam.items():
            f.write('+parameter+{}.{}={}\n'.format(self.toplevel, key, self._param_value_str(value, '"')))
        for id in incdirs:
            f.write("+incdir+" + id+'\n')
        timescale = self.tool_options.get('timescale')
        if timescale:
            with open(os.path.join(self.work_root, 'timescale.v'), 'w') as tsfile:
                tsfile.write("`timescale {}\n".format(timescale))
            f.write('timescale.v\n')
        for src_file in src_files:
            if src_file.file_type in ["verilogSource",
		                      "verilogSource-95",
		                      "verilogSource-2001",
		                      "verilogSource-2005",
                                      "systemVerilogSource",
			              "systemVerilogSource-3.0",
			              "systemVerilogSource-3.1",
			              "systemVerilogSource-3.1a"]:
                f.write(src_file.name+'\n')
            elif src_file.file_type == 'user':
                pass
            else:
                _s = "{} has unknown file type '{}'"
                logger.warning(_s.format(src_file.name, src_file.file_type))

        f.close()

        with open(os.path.join(self.work_root, 'Makefile'), 'w') as f:

            f.write("TARGET           := {}\n".format(self.name))
            _vpi_modules = ' '.join([m['name']+'.vpi' for m in self.vpi_modules])
            if _vpi_modules:
                f.write("VPI_MODULES      := {}\n".format(_vpi_modules))
            f.write("TOPLEVEL         := {}\n".format(self.toplevel))
            f.write("IVERILOG_OPTIONS := {}\n".format(' '.join(self.tool_options.get('iverilog_options', []))))
            if self.plusarg:
                plusargs = []
                for key, value in self.plusarg.items():
                    plusargs += ['+{}={}'.format(key, self._param_value_str(value))]
                f.write("EXTRA_OPTIONS    ?= {}\n".format(' '.join(plusargs)))

            f.write(self.MAKEFILE_TEMPLATE)

            for vpi_module in self.vpi_modules:
                _incs = ['-I' + s for s in vpi_module['include_dirs']]
                _libs = ['-l'+l for l in vpi_module['libs']]
                _srcs = vpi_module['src_files']
                f.write(self.VPI_MAKE_SECTION.format(name = vpi_module['name'],
                                                     libs = ' '.join(_libs),
                                                     incs = ' '.join(_incs),
                                                     srcs = ' '.join(_srcs)))

    def run_main(self):
        args = ['run']

        # Set plusargs
        if self.plusarg:
            plusargs = []
            for key, value in self.plusarg.items():
                plusargs += ['+{}={}'.format(key, self._param_value_str(value))]
            args.append('EXTRA_OPTIONS='+' '.join(plusargs))

        self._run_tool('make', args)



import os
import logging

from edalize.edatool import Edatool

logger = logging.getLogger(__name__)

MAKE_HEADER ="""#Generated by Edalize
ifndef MODEL_TECH
$(error Environment variable MODEL_TECH was not found. It should be set to <modelsim install path>/bin)
endif

CC ?= gcc
CFLAGS := -c -std=c99 -fPIC -fno-stack-protector -g -m32

LD ?= ld
LDFLAGS := -shared -E -melf_i386

RM ?= rm
INCS := -I$(MODEL_TECH)/../include

VSIM ?= $(MODEL_TECH)/vsim

TOPLEVEL      := {toplevel}
VPI_MODULES   := {modules}
PARAMETERS    ?= {parameters}
PLUSARGS      ?= {plusargs}
VSIM_OPTIONS  ?= {vsim_options}
EXTRA_OPTIONS ?= $(VSIM_OPTIONS) $(addprefix -g,$(PARAMETERS)) $(addprefix +,$(PLUSARGS))

all: work $(VPI_MODULES)

run: work $(VPI_MODULES)
	$(VSIM) -do "run -all" -c $(addprefix -pli ,$(VPI_MODULES)) $(TOPLEVEL) $(EXTRA_OPTIONS)

run-gui: work $(VPI_MODULES)
	$(VSIM) -gui $(addprefix -pli ,$(VPI_MODULES)) $(TOPLEVEL) $(EXTRA_OPTIONS)

work:
	$(VSIM) -c -do "do edalize_main.tcl; exit"

clean: {clean_targets}
"""

VPI_MAKE_SECTION = """
{name}_OBJS := {objs}
{name}_LIBS := {libs}
{name}_INCS := $(INCS) {incs}

$({name}_OBJS): %.o : %.c
	$(CC) $(CFLAGS) $({name}_INCS) -o $@ $<

{name}: $({name}_OBJS)
	$(LD) $(LDFLAGS) -o $@ $? $({name}_LIBS)

clean_{name}:
	$(RM) $({name}_OBJS) {name}
"""

class Modelsim(Edatool):

    tool_options = {'lists' : {'vlog_options' : 'String',
                               'vsim_options' : 'String'}}

    argtypes = ['plusarg', 'vlogdefine', 'vlogparam']

    def _write_build_rtl_tcl_file(self, tcl_main):
        tcl_build_rtl  = open(os.path.join(self.work_root, "edalize_build_rtl.tcl"), 'w')

        (src_files, incdirs) = self._get_fileset_files()
        vlog_include_dirs = ['+incdir+'+d.replace('\\','/') for d in incdirs]

        libs = []
        for f in src_files:
            if not f.logical_name:
                f.logical_name = 'work'
            if not f.logical_name in libs:
                tcl_build_rtl.write("vlib {}\n".format(f.logical_name))
                libs.append(f.logical_name)
            if f.file_type.startswith("verilogSource") or \
               f.file_type.startswith("systemVerilogSource"):
                cmd = 'vlog'
                args = []

                args += self.tool_options.get('vlog_options', [])

                for k, v in self.vlogdefine.items():
                    args += ['+define+{}={}'.format(k,self._param_value_str(v))]

                if f.file_type.startswith("systemVerilogSource"):
                    args += ['-sv']
                args += vlog_include_dirs
            elif f.file_type.startswith("vhdlSource"):
                cmd = 'vcom'
                if f.file_type.endswith("-87"):
                    args = ['-87']
                if f.file_type.endswith("-93"):
                    args = ['-93']
                if f.file_type.endswith("-2008"):
                    args = ['-2008']
                else:
                    args = []
            elif f.file_type == 'tclSource':
                cmd = None
                tcl_main.write("do {}\n".format(f.name))
            elif f.file_type == 'user':
                cmd = None
            else:
                _s = "{} has unknown file type '{}'"
                logger.warning(_s.format(f.name, f.file_type))
                cmd = None
            if cmd:
                args += ['-quiet']
                args += ['-work', f.logical_name]
                args += [f.name.replace('\\','/')]
                tcl_build_rtl.write("{} {}\n".format(cmd, ' '.join(args)))

    def _write_makefile(self):
        vpi_make = open(os.path.join(self.work_root, "Makefile"), 'w')
        _parameters = []
        for key, value in self.vlogparam.items():
            _parameters += ['{}={}'.format(key, self._param_value_str(value))]
        _plusargs = []
        for key, value in self.plusarg.items():
            _plusargs += ['{}={}'.format(key, self._param_value_str(value))]

        _vsim_options = self.tool_options.get('vsim_options', [])

        _modules = [m['name'] for m in self.vpi_modules]
        _clean_targets = ' '.join(["clean_"+m for m in _modules])
        _s = MAKE_HEADER.format(toplevel = self.toplevel,
                                parameters = ' '.join(_parameters),
                                plusargs = ' '.join(_plusargs),
                                vsim_options = ' '.join(_vsim_options),
                                modules = ' '.join(_modules),
                                clean_targets = _clean_targets)
        vpi_make.write(_s)

        for vpi_module in self.vpi_modules:
            _name = vpi_module['name']
            _objs = [os.path.splitext(s)[0]+'.o' for s in vpi_module['src_files']]
            _libs = ['-l'+l for l in vpi_module['libs']]
            _incs = ['-I'+d for d in vpi_module['include_dirs']]
            _s = VPI_MAKE_SECTION.format(name=_name,
                                         objs=' '.join(_objs),
                                         libs=' '.join(_libs),
                                         incs=' '.join(_incs))
            vpi_make.write(_s)

        vpi_make.close()

    def configure_main(self):
        tcl_main = open(os.path.join(self.work_root, "edalize_main.tcl"), 'w')
        tcl_main.write("do edalize_build_rtl.tcl\n")

        self._write_build_rtl_tcl_file(tcl_main)
        self._write_makefile()
        tcl_main.close()

    def run_main(self):
        args = ['run']

        # Set plusargs
        if self.plusarg:
            plusargs = []
            for key, value in self.plusarg.items():
                plusargs += ['{}={}'.format(key, self._param_value_str(value))]
            args.append('PLUSARGS='+' '.join(plusargs))

        self._run_tool('make', args)



import argparse
from collections import OrderedDict
import os
import subprocess
import logging
import sys
import yaml
from jinja2 import Environment, PackageLoader

logger = logging.getLogger(__name__)

# Jinja2 tests and filters, available in all templates
def jinja_is_verilog_file(f):
    return f.file_type.startswith('verilogSource')

def jinja_is_system_verilog_file(f):
    return f.file_type.startswith('systemVerilogSource')

def jinja_is_vhdl_file(f):
    return f.file_type.startswith('vhdlSource')

def jinja_filter_param_value_str(value, str_quote_style=""):
    """ Convert a parameter value to string suitable to be passed to an EDA tool

    Rules:
    - Booleans are represented as 0/1
    - Strings are either passed through or enclosed in the characters specified
      in str_quote_style (e.g. '"' or '\\"')
    - Everything else (including int, float, etc.) are converted using the str()
      function.
    """
    if type(value) == bool:
        if (value) == True:
            return '1'
        else:
            return '0'
    elif type(value) == str:
        return str_quote_style + str(value) + str_quote_style
    else:
        return str(value)


class FileAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        path = os.path.expandvars(values[0])
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        setattr(namespace, self.dest, [path])

class Edatool(object):

    def __init__(self, eda_api_file, work_root=None):
        _tool_name = self.__class__.__name__.lower()

        eda_api = yaml.load(open(eda_api_file))

        if not eda_api:
            raise RuntimeError("Failed to parse " + eda_api_file)

        try:
            self.name = eda_api['name']
        except KeyError:
            raise RuntimeError("Missing required parameter 'name'")

        self.tool_options = eda_api.get('tool_options', {}).get(_tool_name, {})

        self.files       = eda_api.get('files', [])
        self.toplevel    = eda_api.get('toplevel', [])
        self.vpi_modules = eda_api.get('vpi', [])

        self.hooks       = eda_api.get('hooks', {})
        self.parameters  = eda_api.get('parameters', {})

        self.work_root = work_root or os.path.abspath(os.path.dirname(eda_api_file))
        self.env = os.environ.copy()

        self.env['WORK_ROOT'] = self.work_root

        self.plusarg     = OrderedDict()
        self.vlogparam   = OrderedDict()
        self.vlogdefine  = OrderedDict()
        self.generic     = OrderedDict()
        self.cmdlinearg  = OrderedDict()
        self.parsed_args = False

        self.jinja_env = Environment(
            loader = PackageLoader(__package__, 'templates'),
            trim_blocks = True,
            lstrip_blocks = True,
        )
        self.jinja_env.tests['verilog_file'] = jinja_is_verilog_file
        self.jinja_env.tests['system_verilog_file'] = jinja_is_system_verilog_file
        self.jinja_env.tests['vhdl_file'] = jinja_is_vhdl_file
        self.jinja_env.filters['param_value_str'] = jinja_filter_param_value_str


    def configure(self, args):
        logger.info("Setting up project")
        self.configure_pre(args)
        self.configure_main()
        self.configure_post()

    def configure_pre(self, args):
        self.parse_args(args, self.argtypes)

    def configure_main(self):
        pass

    def configure_post(self):
        pass

    def build(self):
        self.build_pre()
        self.build_main()
        self.build_post()

    def build_pre(self):
        if 'pre_build' in self.hooks:
            self._run_scripts(self.hooks['pre_build'])

    def build_main(self):
        logger.info("Building");
        self._run_tool('make')

    def build_post(self):
        if 'post_build' in self.hooks:
            self._run_scripts(self.hooks['post_build'])

    def run(self, args):
        logger.info("Running")
        self.run_pre(args)
        self.run_main()
        self.run_post()

    def run_pre(self, args):
        self.parse_args(args, self.argtypes)
        if 'pre_run' in self.hooks:
            self._run_scripts(self.hooks['pre_run'])

    def run_main(self):
        pass

    def run_post(self):
        if 'post_run' in self.hooks:
            self._run_scripts(self.hooks['post_run'])

    def parse_args(self, args, paramtypes):
        if self.parsed_args:
            return
        typedict = {'bool' : {'action' : 'store_true'},
                    'file' : {'type' : str , 'nargs' : 1, 'action' : FileAction},
                    'int'  : {'type' : int , 'nargs' : 1},
                    'str'  : {'type' : str , 'nargs' : 1},
                    }
        progname = os.path.basename(sys.argv[0]) + ' run {}'.format(self.name)

        parser = argparse.ArgumentParser(prog = progname,
                                         conflict_handler='resolve')
        param_groups = {}
        _descr = {'plusarg'    : 'Verilog plusargs (Run-time option)',
                  'vlogparam'  : 'Verilog parameters (Compile-time option)',
                  'vlogdefine' : 'Verilog defines (Compile-time global symbol)',
                  'generic'    : 'VHDL generic (Run-time option)',
                  'cmdlinearg' : 'Command-line arguments (Run-time option)'}
        param_type_map = {}

        for name, param in self.parameters.items():
            _description = param.get('description', "No description")
            _paramtype = param['paramtype']
            if _paramtype in paramtypes:
                if not _paramtype in param_groups:
                    param_groups[_paramtype] = \
                    parser.add_argument_group(_descr[_paramtype])

                default = None
                if param.get('default'):
                    try:
                        default = [typedict[param['datatype']]['type'](param['default'])]
                    except KeyError as e:
                        pass
                try:
                    param_groups[_paramtype].add_argument('--'+name,
                                                               help=_description,
                                                               default=default,
                                                               **typedict[param['datatype']])
                except KeyError as e:
                    raise RuntimeError("Invalid data type {} for parameter '{}'".format(str(e),
                                                                                        name))
                param_type_map[name.replace('-','_')] = _paramtype
        #Parse arguments
        for key,value in sorted(vars(parser.parse_args(args)).items()):

            paramtype = param_type_map[key]
            if value is None:
                continue

            if type(value) == bool:
                _value = value
            else:
                _value = value[0]

            getattr(self, paramtype)[key] = _value
        self.parsed_args = True

    def render_template(self, template_file, target_file, template_vars = {}):
        """
        Render a Jinja2 template for the backend

        The template file is expected in the directory templates/BACKEND_NAME.
        """
        template_dir = str(self.__class__.__name__).lower()
        template = self.jinja_env.get_template(os.path.join(template_dir, template_file))
        file_path = os.path.join(self.work_root, target_file)
        with open(file_path, 'w') as f:
            f.write(template.render(template_vars))

    def _get_fileset_files(self, force_slash=False):
        class File:
            def __init__(self, name, file_type, logical_name):
                self.name         = name
                self.file_type    = file_type
                self.logical_name = logical_name
        incdirs = []
        src_files = []
        for f in self.files:
            if 'is_include_file' in f and f['is_include_file']:
                _incdir = os.path.dirname(f['name']) or '.'
                if force_slash:
                    _incdir = _incdir.replace('\\', '/')
                if not _incdir in incdirs:
                    incdirs.append(_incdir)
            else:
                _name = f['name']
                if force_slash:
                    _name = _name.replace('\\', '/')
                file_type = f.get('file_type', '')
                logical_name = f.get('logical_name', '')
                src_files.append(File(_name,
                                      file_type,
                                      logical_name))
        return (src_files, incdirs)

    def _param_value_str(self, param_value, str_quote_style=""):
        return jinja_filter_param_value_str(param_value, str_quote_style)

    def _run_scripts(self, scripts):
        for script in scripts:
            _env = self.env.copy()
            if 'env' in script:
                _env.update(script['env'])
            logger.info("Running " + script['name'])
            try:
                subprocess.check_call(script['cmd'],
                                      cwd = self.work_root,
                                      env = _env)
            except subprocess.CalledProcessError as e:
                msg = "'{}' exited with error code {}"
                raise RuntimeError(msg.format(script['name'], e.returncode))

    def _run_tool(self, cmd, args=[]):
        logger.debug("Running " + cmd)
        logger.debug("args  : " + ' '.join(args))

        try:
            subprocess.check_call([cmd] + args,
                                  cwd = self.work_root,
                                  stdin=subprocess.PIPE),
        except FileNotFoundError:
            _s = "Command '{}' not found. Make sure it is in $PATH"
            raise RuntimeError(_s.format(cmd))
        except subprocess.CalledProcessError:
            _s = "'{}' exited with an error code"
            raise RuntimeError(_s.format(cmd))



import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "fusesoc",
    packages=['fusesoc',
              'fusesoc.capi1',
              'fusesoc.capi2',
              'edalize',
              'fusesoc.provider'],
    package_data = {'edalize' : [
        'templates/icestorm/icestorm-makefile.j2',
        'templates/spyglass/Makefile.j2',
        'templates/spyglass/spyglass-project.prj.j2',
        'templates/spyglass/spyglass-run-goal.tcl.j2',
        'templates/vivado/vivado-makefile.j2',
        'templates/vivado/vivado-program.tcl.j2',
        'templates/vivado/vivado-project.tcl.j2',
        'templates/vivado/vivado-run.tcl.j2',
    ]},
    use_scm_version = {
        "relative_to": __file__,
        "write_to": "fusesoc/version.py",
    },
    author = "Olof Kindgren",
    author_email = "olof.kindgren@gmail.com",
    description = ("FuseSoC is a package manager and a set of build tools for HDL (Hardware Description Language) code."),
    license = "GPLv3",
    keywords = ["VHDL", "verilog", "hdl", "rtl", "synthesis", "FPGA", "simulation", "Xilinx", "Altera"],
    url = "https://github.com/olofk/fusesoc",
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    entry_points={
        'console_scripts': [
            'fusesoc = fusesoc.main:main'
        ]
    },
    setup_requires=[
        'setuptools_scm',
    ],
    install_requires=[
        'ipyxact>=0.2.3',
        'pyparsing',
        'pytest>=3.3.0',
        'pyyaml',
        'simplesat>=0.8.0',
        'Jinja2>=2.8',
    ],
)



import multiprocessing
import os
import logging

from edalize.edatool import Edatool

logger = logging.getLogger(__name__)

CONFIG_MK_TEMPLATE = """#Auto generated by Edalize

TOP_MODULE        := {top_module}
VC_FILE           := {vc_file}
VERILATOR_OPTIONS := {verilator_options}
"""

MAKEFILE_TEMPLATE = """#Auto generated by Edalize

include config.mk

#Assume a local installation if VERILATOR_ROOT is set
ifeq ($(VERILATOR_ROOT),)
VERILATOR ?= verilator
else
VERILATOR ?= $(VERILATOR_ROOT)/bin/verilator
endif

V$(TOP_MODULE): V$(TOP_MODULE).mk
	$(MAKE) -f $<

V$(TOP_MODULE).mk:
	$(VERILATOR) -f $(VC_FILE) $(VERILATOR_OPTIONS)
"""

class Verilator(Edatool):

    tool_options = {'members' : {'mode' : 'String',
                                 'cli_parser' : 'String'},
                    'lists'   : {'libs' : 'String',
                                 'verilator_options' : 'String'}}

    argtypes = ['cmdlinearg', 'plusarg', 'vlogdefine', 'vlogparam']

    def _managed_parser(self):
        return not 'cli_parser' in self.tool_options or self.tool_options['cli_parser'] == 'managed'

    def configure_pre(self, args):
        if self._managed_parser():
            self.parse_args(args, self.argtypes)

    def configure_main(self):
        if not self.toplevel:
            raise RuntimeError("'" + self.name + "' miss a mandatory parameter 'top_module'")

        self._write_config_files()

    def _write_config_files(self):
        #Future improvement: Separate include directories of c and verilog files
        incdirs = set()
        src_files = []

        (src_files, incdirs) = self._get_fileset_files()

        self.verilator_file = self.name + '.vc'

        with open(os.path.join(self.work_root,self.verilator_file),'w') as f:
            f.write('--Mdir .\n')
            modes = ['sc', 'cc', 'lint-only']

            #Default to cc mode if not specified
            if not 'mode' in self.tool_options:
                self.tool_options['mode'] = 'cc'

            if self.tool_options['mode'] in modes:
                f.write('--'+self.tool_options['mode']+'\n')
            else:
                _s = "Illegal verilator mode {}. Allowed values are {}"
                raise RuntimeError(_s.format(self.tool_options['mode'],
                                             ', '.join(modes)))
            if 'libs' in self.tool_options:
                for lib in self.tool_options['libs']:
                    f.write('-LDFLAGS {}\n'.format(lib))
            for include_dir in incdirs:
                f.write("+incdir+" + include_dir + '\n')
                f.write("-CFLAGS -I{}\n".format(include_dir))
            opt_c_files = []
            for src_file in src_files:
                if src_file.file_type.startswith("systemVerilogSource") or src_file.file_type.startswith("verilogSource"):
                    f.write(src_file.name + '\n')
                elif src_file.file_type in ['cppSource', 'systemCSource', 'cSource']:
                    opt_c_files.append(src_file.name)
                elif src_file.file_type in ['user']:
                    pass
            f.write('--top-module {}\n'.format(self.toplevel))
            f.write('--exe\n')
            f.write('\n'.join(opt_c_files))
            f.write('\n')
            f.write(''.join(['-G{}={}\n'.format(key, self._param_value_str(value)) for key, value in self.vlogparam.items()]))
            f.write(''.join(['-D{}={}\n'.format(key, self._param_value_str(value)) for key, value in self.vlogdefine.items()]))

        with open(os.path.join(self.work_root, 'Makefile'), 'w') as makefile:
            makefile.write(MAKEFILE_TEMPLATE)

        if 'verilator_options' in self.tool_options:
            verilator_options = ' '.join(self.tool_options['verilator_options'])
        else:
            verilator_options = ''
        with open(os.path.join(self.work_root, 'config.mk'), 'w') as config_mk:
            config_mk.write(CONFIG_MK_TEMPLATE.format(
                top_module        = self.toplevel,
                vc_file           = self.verilator_file,
                verilator_options = verilator_options))

    def build_main(self):
        logger.info("Building simulation model")

        # Do parallel builds with <number of cpus> * 2 jobs.
        make_job_count = multiprocessing.cpu_count() * 2
        args = ['-j', str(make_job_count)]

        if self.tool_options['mode'] == 'lint-only':
            args.append('V'+self.toplevel+'.mk')
        _s = os.path.join(self.work_root, 'verilator.{}.log')
        self._run_tool('make', args)

    def run(self, args):
        if self.tool_options['mode'] == 'lint-only':
            return
        if self._managed_parser():
            self.parse_args(args, self.argtypes)

            _args = []
            for key, value in self.plusarg.items():
                _args += ['+{}={}'.format(key, self._param_value_str(value))]
            for key, value in self.cmdlinearg.items():
                _args += ['--{}={}'.format(key, self._param_value_str(value))]
        else:
            _args = args

        if 'pre_run' in self.hooks:
            self._run_scripts(self.hooks['pre_run'])

        logger.info("Running simulation")
        self._run_tool('./V' + self.toplevel, _args)

        self.run_post()



import os.path
import tempfile
import yaml

from edalize import get_edatool

tests_dir = os.path.dirname(__file__)

def compare_files(ref_dir, work_root, files):
    import os.path

    for f in files:
        reference_file = os.path.join(ref_dir, f)
        generated_file = os.path.join(work_root, f)

        assert os.path.exists(generated_file)

        with open(reference_file) as fref, open(generated_file) as fgen:
            assert fref.read() == fgen.read(), f

def param_gen(paramtypes):
    args = []
    defs = {}
    for paramtype in paramtypes:
        for datatype in ['bool', 'int', 'str']:
            _arg = '--{}_{}'.format(paramtype, datatype)
            if datatype == 'int':
                _arg += '=42'
            elif datatype == 'str':
                _arg += '=hello'
            args.append(_arg)
            defs[paramtype+'_'+datatype] = {
                'datatype'    : datatype,
                'default'     : '',
                'description' : '',
                'paramtype'   : paramtype}
    return (defs, args)

def setup_backend_minimal(name, tool, files):
    os.environ['PATH'] = os.path.join(tests_dir, 'mock_commands')+':'+os.environ['PATH']

    work_root = tempfile.mkdtemp(prefix=tool+'_')

    eda_api_file = os.path.join(work_root, name+'.eda.yml')
    with open(eda_api_file,'w') as f:
        f.write(yaml.dump({'name'         : name,
                           'files'        : files,
                           'toplevel'     : 'top_module',
                           }))
    return (get_edatool(tool)(eda_api_file=eda_api_file), work_root)

def setup_backend(paramtypes, name, tool, tool_options, use_vpi=False):
    os.environ['PATH'] = os.path.join(tests_dir, 'mock_commands')+':'+os.environ['PATH']
    (parameters, args) = param_gen(paramtypes)

    work_root = tempfile.mkdtemp(prefix=tool+'_')

    _vpi = []
    if use_vpi:
        _vpi = vpi
        for v in vpi:
            for f in v['src_files']:
                _f = os.path.join(work_root, f)
                if not os.path.exists(os.path.dirname(_f)):
                    os.makedirs(os.path.dirname(_f))
                with open(_f, 'a'):
                    os.utime(_f, None)

    eda_api_file = os.path.join(work_root, name+'.eda.yml')
    with open(eda_api_file,'w') as f:
        f.write(yaml.dump({'name'         : name,
                           'files'        : files,
                           'parameters'   : parameters,
                           'tool_options' : {tool : tool_options},
                           'toplevel'     : 'top_module',
                           'vpi'          :  _vpi}))

    return (get_edatool(tool)(eda_api_file=eda_api_file), args, work_root)

files = [
    {'name' : 'qip_file.qip' , 'file_type' : 'QIP'},
    {'name' : 'qsys_file'    , 'file_type' : 'QSYS'},
    {'name' : 'sdc_file'     , 'file_type' : 'SDC'},
    {'name' : 'bmm_file'     , 'file_type' : 'BMM'},
    {'name' : 'sv_file.sv'   , 'file_type' : 'systemVerilogSource'},
    {'name' : 'pcf_file.pcf' , 'file_type' : 'PCF'},
    {'name' : 'ucf_file.ucf' , 'file_type' : 'UCF'},
    {'name' : 'user_file'    , 'file_type' : 'user'},
    {'name' : 'tcl_file.tcl' , 'file_type' : 'tclSource'},
    {'name' : 'vlog_file.v'  , 'file_type' : 'verilogSource'},
    {'name' : 'vlog05_file.v', 'file_type' : 'verilogSource-2005'},
    {'name' : 'vlog_incfile' , 'file_type' : 'verilogSource', 'is_include_file' : True},
    {'name' : 'vhdl_file.vhd', 'file_type' : 'vhdlSource'},
    {'name' : 'vhdl_lfile'   , 'file_type' : 'vhdlSource', 'logical_name' : 'libx'},
    {'name' : 'vhdl2008_file', 'file_type' : 'vhdlSource-2008'},
    {'name' : 'xci_file.xci' , 'file_type' : 'xci'},
    {'name' : 'xdc_file.xdc' , 'file_type' : 'xdc'}
]

vpi = [
    {'src_files': ['src/vpi_1/f1',
                   'src/vpi_1/f3'],
     'include_dirs': ['src/vpi_1/'],
     'libs': ['some_lib'],
     'name': 'vpi1'},
    {'src_files': ['src/vpi_2/f4'],
     'include_dirs': [],
     'libs': [],
     'name': 'vpi2'}]
    



import os
import logging

from edalize.edatool import Edatool

logger = logging.getLogger(__name__)

class Isim(Edatool):

    tool_options = {'lists' : {'fuse_options' : 'String',
                               'isim_options' : 'String'}}

    argtypes = ['plusarg', 'vlogdefine', 'vlogparam']

    MAKEFILE_TEMPLATE="""#Auto generated by Edalize
include config.mk

all: $(TARGET)

$(TARGET):
	fuse $(TOPLEVEL) -prj $(TARGET).prj -o $(TARGET) $(VLOG_DEFINES) $(VLOG_INCLUDES) $(VLOG_PARAMS) $(FUSE_OPTIONS)

run: $(TARGET)
	./$(TARGET) -tclbatch run_$(TARGET).tcl $(ISIM_OPTIONS) $(EXTRA_OPTIONS)

run-gui: $(TARGET)
	./$(TARGET) -gui $(ISIM_OPTIONS) $(EXTRA_OPTIONS)
"""

    CONFIG_MK_TEMPLATE = """#Auto generated by Edalize
TARGET        = {target}
TOPLEVEL      = {toplevel}

VLOG_DEFINES  = {vlog_defines}
VLOG_INCLUDES = {vlog_includes}
VLOG_PARAMS   = {vlog_params}

FUSE_OPTIONS  =	{fuse_options}
ISIM_OPTIONS  =	{isim_options}

EXTRA_OPTIONS ?= {extra_options}
"""

    RUN_TCL_TEMPLATE = """#Auto generated by Edalize
wave log -r /
run all
quit
"""

    def configure_main(self):
        #Check if any VPI modules are present and display warning
        if len(self.vpi_modules) > 0:
            modules = [m['name'] for m in self.vpi_modules]
            logger.error('VPI modules not supported by Isim: %s' % ', '.join(modules))

        with open(os.path.join(self.work_root, self.name+'.prj'),'w') as f:
            (src_files, self.incdirs) = self._get_fileset_files()

            for src_file in src_files:
                prefix = ""
                logical_name = ""
                if src_file.file_type in [
                        "verilogSource",
                        "verilogSource-95",
                        "verilogSource-2001"]:
                    prefix = 'verilog'
                elif src_file.file_type.startswith("vhdlSource"):
                    prefix = 'vhdl'
                    if src_file.logical_name:
                        logical_name = src_file.logical_name + ' '
                elif src_file.file_type in ["systemVerilogSource",
                                            "systemVerilogSource-3.0",
                                            "systemVerilogSource-3.1",
                                            "systemVerilogSource-3.1a",
                                            "verilogSource-2005"]:
                    prefix = 'sv'
                elif src_file.file_type in ["user"]:
                    pass
                else:
                    _s = "{} has unknown file type '{}'"
                    logger.warning(_s.format(src_file.name,
                               src_file.file_type))
                if prefix:
                    f.write('{} work {}{}\n'.format(prefix, logical_name, src_file.name))

        with open(os.path.join(self.work_root, 'run_'+self.name+'.tcl'),'w') as f:
            f.write(self.RUN_TCL_TEMPLATE)

        with open(os.path.join(self.work_root, 'Makefile'),'w') as f:
            f.write(self.MAKEFILE_TEMPLATE)

        with open(os.path.join(self.work_root, 'config.mk'),'w') as f:
            vlog_defines  = ' '.join(['--define {}={}'.format(k, self._param_value_str(v)) for k,v, in self.vlogdefine.items()])
            vlog_includes = ' '.join(['-i '+k for k in self.incdirs])
            vlog_params   = ' '.join(['--generic_top {}={}'.format(k, self._param_value_str(v)) for k,v, in self.vlogparam.items()])
            fuse_options = ' '.join(self.tool_options.get('fuse_options', []))
            isim_options = ' '.join(self.tool_options.get('isim_options', []))

            _s = '-testplusarg {}={}'
            extra_options = ' '.join([_s.format(k, self._param_value_str(v)) for k,v in self.plusarg.items()])
            f.write(self.CONFIG_MK_TEMPLATE.format(target        = self.name,
                                                   toplevel      = self.toplevel,
                                                   vlog_defines  = vlog_defines,
                                                   vlog_includes = vlog_includes,
                                                   vlog_params   = vlog_params,
                                                   fuse_options  = fuse_options,
                                                   isim_options  = isim_options,
                                                   extra_options = extra_options))

    def run_main(self):
        args = ['run']
        # Plusargs
        if self.plusarg:
            _s = '-testplusarg {}={}'
            args.append('EXTRA_OPTIONS='+' '.join([_s.format(k, self._param_value_str(v)) for k,v in self.plusarg.items()]))
        self._run_tool('make', args)



import pytest

from fusesoc.vlnv import Vlnv

def vlnv_tuple(vlnv):
    return (vlnv.vendor, vlnv.library, vlnv.name, vlnv.version, vlnv.revision)

#VLNV tests without revision
def test_full_vlnv():
    assert vlnv_tuple(Vlnv("librecores.org:peripherals:uart16550:1.5")) == \
    ('librecores.org', 'peripherals', 'uart16550', '1.5', 0)

def test_full_vlnv_no_version():
    assert vlnv_tuple(Vlnv("librecores.org:peripherals:uart16550")) == \
    ('librecores.org', 'peripherals', 'uart16550', '0', 0)

def test_name_only_vlnv():
    assert vlnv_tuple(Vlnv("::uart16550")) == \
    ('', '', 'uart16550', '0', 0)
    assert vlnv_tuple(Vlnv("::uart16550:")) == \
    ('', '', 'uart16550', '0', 0)
    
def test_name_version_vlnv():
    assert vlnv_tuple(Vlnv("::uart16550:1.5")) == \
    ('', '', 'uart16550', '1.5', 0)

#VLNV tests with revision
def test_full_vlnv_revision():
    assert vlnv_tuple(Vlnv("librecores.org:peripherals:uart16550:1.5-r5")) == \
    ('librecores.org', 'peripherals', 'uart16550', '1.5', 5)

def test_name_only_vlnv_revision():
    assert vlnv_tuple(Vlnv("::uart16550")) == \
    ('', '', 'uart16550', '0', 0)
    assert vlnv_tuple(Vlnv("::uart16550:")) == \
    ('', '', 'uart16550', '0', 0)
    
def test_name_version_vlnv():
    assert vlnv_tuple(Vlnv("::uart16550:1.5")) == \
    ('', '', 'uart16550', '1.5', 0)

#Tests for legacy naming scheme
def test_name_version_legacy():
    assert vlnv_tuple(Vlnv("uart16550-1.5")) == \
    ('', '', 'uart16550', '1.5', 0)

def test_name_with_dash_version_legacy():
    assert vlnv_tuple(Vlnv("wb-axi-1.5")) == \
    ('', '', 'wb-axi', '1.5', 0)

def test_name_only_legacy():
    assert vlnv_tuple(Vlnv("uart16550")) == \
    ('', '', 'uart16550', '0', 0)

def test_name_with_dash_only_legacy():
    assert vlnv_tuple(Vlnv("wb-axi")) == \
    ('', '', 'wb-axi', '0', 0)

def test_name_version_revision_legacy():
    assert vlnv_tuple(Vlnv("uart16550-1.5-r2")) == \
    ('', '', 'uart16550', '1.5', 2)

def test_name_revision_legacy():
    assert vlnv_tuple(Vlnv("uart16550-r2")) == \
    ('', '', 'uart16550', '0', 2)

